"""Example file that will have different formatting based on pyink configs."""

from very.long.package.path.my_org.my_very_long_project_name.awesome_backend.core_framework.util import my_long_module_name


def function():
 pass
