"""Docstring quotes are not counted."""

_multiline = """Multi-line string quotes are not counted either."""

_single = 'Single'
_another_single = 'Another Single'
_double = "Double"
_fstring_single = f'fstring single {4 + 2}'
_nested_fstring_double = f"nested fstring {f'double {4 + 2}'}"
